#!/usr/bin/env python2

import getopt
import os
import sys
import json
import string
import socket
import logging
from contextlib import closing

# Add own packages to top of the list to avoid package name clashes.
sys.path.insert(1, os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'dist', 'lib', 'multiplatform', 'python'))

import app.servicecontrol
import app.landscape
import app.servicemanifest
import app.options
import app.util
import app.logutil
import app.colors

def verifyUser():
    if not app.util.isSuperUser():
        sys.stderr.write(app.colors.Color.red("Script must be run as root\n"))
        sys.exit(1)

def verifyDeploymentLeftovers():
    services = app.landscape.Landscape.getSite().getLocalMachine().getServices()
    configuredServiceNames = [s.getName() for s in services]

    # Check if there is any deployed service which is not part of the configured
    # system landscape anymore.
    deployedServiceNames = app.servicecontrol.DeploymentManager.getAllServicesMarkedAsDeployed()
    leftovers = [serviceName for serviceName in deployedServiceNames
                 if serviceName not in configuredServiceNames]

    if len(leftovers):
        sys.stderr.write(app.colors.Color.red(
            "The following services are currently deployed, but are not part of the \n"+
            "configuration any more: %s\n\n"  % ", ".join(leftovers) +
            "Please re-add them to your configuration and then undeploy them\n\n"))

def awaitStartupDependencies(logger, service=None):
    site = app.landscape.Landscape.getSite()

    if service:
        controller = service.getController(logger)
    else:
        controller = app.landscape.Landscape.getSite().getLocalMachine().getController(logger)

    action = controller.createAwaitStartupDependenciesAction()

    if not action.execute():
        sys.exit(1)

def checkPorts(logger, service=None):
    site = app.landscape.Landscape.getSite()

    if service:
        controller = service.getController(logger)
    else:
        controller = app.landscape.Landscape.getSite().getLocalMachine().getController(logger)

    action = controller.createPortAvailabilityCheckAction()

    if not action.execute():
        sys.exit(1)

def verifyEnvironment(logger, service=None):
    site = app.landscape.Landscape.getSite()

    if service:
        controller = service.getController(logger)
    else:
        controller = app.landscape.Landscape.getSite().getLocalMachine().getController(logger)

    action = controller.createEnvironmentVerificationCheckAction()

    if not action.execute():
        sys.exit(1)

def deploy(logger, service=None):
    site = app.landscape.Landscape.getSite()

    if service:
        controller = service.getController(logger)
    else:
        controller = app.landscape.Landscape.getSite().getLocalMachine().getController(logger)

    # Verify the environment. Although this is done implicitly during
    # the deployment process, we do it here in order to get clear output.
    if app.servicecontrol.ServiceControlGlobalFlags.verificationEnabled:
        if not controller.createEnvironmentVerificationCheckAction().execute():
            sys.exit(1)

    # Note: Before we can deploy, we need to undeploy first.
    deployAction = controller.createDeployAction()
    undeployAction = controller.createUndeployAction() \
        .filterDeployedServices()        # Do not undeploy what is not deployed

    # Seek confirmation for both first.
    if app.servicecontrol.ServiceControlGlobalFlags.confirmDangerousActions:
        if not undeployAction.confirm():
            sys.exit(3)
        if not deployAction.confirm():
            sys.exit(3)

    undeployAction.execute()
    deployAction.execute()

def undeploy(logger, service=None, forRedeployment=False):
    site = app.landscape.Landscape.getSite()

    if service:
        # Per-service undeployment is not implemented yet - for now, only
        # allow undeploying everything.
        raise Exception("Undeploying a single service is not supported in this release")
    else:
        controller = app.landscape.Landscape.getSite().getLocalMachine().getController(logger)

    action = controller.createUndeployAction()

    if app.servicecontrol.ServiceControlGlobalFlags.confirmDangerousActions:
        if not action.confirm():
            sys.exit(3)

    action.execute()

def bootstrap(logger, service=None, serviceList=None, allowConfirmationPrompt=True):
    site = app.landscape.Landscape.getSite()

    if service:
        action = service.getController(logger).createBootstrapAction()
    else:
        action = app.landscape.Landscape.getSite().getLocalMachine().getController(logger).createBootstrapAction(serviceList)

    if allowConfirmationPrompt and app.servicecontrol.ServiceControlGlobalFlags.confirmDangerousActions:
        if not action.confirm():
            sys.exit(3)

    if not action.execute():
        sys.exit(1)

def migrate(logger, service=None, serviceList=None, allowConfirmationPrompt=True):
    site = app.landscape.Landscape.getSite()

    if service:
        action = service.getController(logger).createMigrateAction()
    else:
        action = app.landscape.Landscape.getSite().getLocalMachine().getController(logger).createMigrateAction(serviceList)

    if allowConfirmationPrompt and app.servicecontrol.ServiceControlGlobalFlags.confirmDangerousActions:
        if not action.confirm():
            sys.exit(3)

    if not action.execute():
        sys.exit(1)

def start(logger, service=None):
    site = app.landscape.Landscape.getSite()

    if service:
        controller = service.getController(logger)
    else:
        controller = app.landscape.Landscape.getSite().getLocalMachine().getController(logger)

    action = controller.createStartAction()

    if not action:
        # Nothing to do
        return

    if app.servicecontrol.ServiceControlGlobalFlags.confirmDangerousActions:
        if not action.confirm():
            sys.exit(3)

    if not action.execute():
        sys.exit(1)

def stop(logger, service=None):
    site = app.landscape.Landscape.getSite()

    if service:
        controller = service.getController(logger)
    else:
        controller = app.landscape.Landscape.getSite().getLocalMachine().getController(logger)

    action = controller.createStopAction()

    if not action:
        # Nothing to do
        return

    if app.servicecontrol.ServiceControlGlobalFlags.confirmDangerousActions:
        if not action.confirm():
            sys.exit(3)

    if not action.execute():
        sys.exit(1)

def reinitialize(logger, service=None):
    site = app.landscape.Landscape.getSite()

    if service:
        controller = service.getController(logger)
    else:
        controller = site.getLocalMachine().getController(logger)

    action = controller.createReinitializeAction()

    if not action:
        return

    if app.servicecontrol.ServiceControlGlobalFlags.confirmDangerousActions:
        if not action.confirm():
            sys.exit(3)

    if not action.execute():
        sys.exit(1)


def applyPermissions(logger, service=None):
    site = app.landscape.Landscape.getSite()

    if service:
        action = service.getController(logger).createApplyPermissionsAction()
    else:
        action = app.landscape.Landscape.getSite().getLocalMachine().getController(logger).createApplyPermissionsAction()

    if not action.execute():
        sys.exit(1)

def dumpEndpoints(logger):
    site = app.landscape.Landscape.getSite()

    # Show diagnostics
    if not site.getController(logger) \
            .createDumpEndpointsAction() \
            .execute():
        sys.exit(1)

def dumpDeploymentDependencies(logger):
    site = app.landscape.Landscape.getSite()

    # Show diagnostics
    if not site.getController(logger) \
            .createDumpDeploymentDependenciesAction() \
            .execute():
        sys.exit(1)

def status(logger, service=None):
    site = app.landscape.Landscape.getSite()

    if service:
        controller = service.getController(logger)
    else:
        controller = app.landscape.Landscape.getSite().getLocalMachine().getController(logger)

    if not controller.createPrintStatusAction().execute():
        sys.exit(1)

def provision(logger, awaitDependencies):
    site = app.landscape.Landscape.getSite()
    localMachine = site.getLocalMachine()

    # If at least one service was deployed already, consider this to be a reprovisioning/upgrade.
    isReprovision = len([s for s in localMachine.getServices() if s.getController(logger).isDeployed()]) > 0

    # 1. Check if all machines this machine depends on are up and running.
    if len(site.getAllMachines()) > 1:
        print "This machine is part of a multi-machine installation. To ensure a smooth operation, these"
        print "machines should be provisioned in the following order:"
        print ""

        index = 0
        for machine in site.getMachineStartupOrder():
            index = index + 1
            print " %d. %s (%s) %s" % (index, machine.getPhysicalName(), machine.getPublicName(), "<- this machine" if machine == localMachine else "")

        print ""
        print "The same order should be followed for (re)starting machines."
        print ""

    machineDependencies = localMachine.getMachineStartupDependencies()
    if machineDependencies:
        print app.colors.Color.yellow("Please make sure that the following machines are up, run the right version of TeamForge,")
        print app.colors.Color.yellow("and are properly configured before proceeding with the installation.")
        print ""
        print " - " + "\n - ".join([m.getPhysicalName() for m in machineDependencies])
        print ""
        print app.colors.Color.yellow("Please also ensure that the AUTO_DATA token (in site-options.conf) is consistent across all machines.")

        if awaitDependencies:
            # Await dependencies.
            if not localMachine.getController(logger).createAwaitStartupDependenciesAction().execute():
                sys.exit(1)

        else:
            # Rely on user to await dependencies.
            if app.servicecontrol.ServiceControlGlobalFlags.confirmDangerousActions:
                print ""
                print "Press any key to continue"
                raw_input()

    # 2. Redeploy.
    deploy(logger)

    # 3. Bootstrap.
    servicesToBootstrap = [s for s in localMachine.getServices()
        if s.getController(logger).getDataStatus() == app.servicecontrol.ServiceStatus.NOT_BOOTSTRAPPED]
    if servicesToBootstrap:
        bootstrap(logger, None, servicesToBootstrap, isReprovision)

    # 4. Migrate - note that it is important (and harmless) to run all migration hooks, all the time.
    # Relying on a service reporting NOT_MIGRATED state may not be reliable in all cases.
    migrate(logger, allowConfirmationPrompt=isReprovision)

    # 5. Start.
    start(logger)

    # 6. All set.
    print ""
    print app.colors.Color.green("Provisioning completed successfully.")
    print ""

def run(logger, awaitDependencies):
    provision(logger, awaitDependencies)

    print app.colors.Color.green("Teamforge is running.")

    while True:
        print ""
        choice = app.util.readFromUser(
            "Press (q) to stop TeamForge, (s) to show the status, or (r) to return to the shell:",
            "s",
            ["q", "s", "r"])

        if choice == "s":
            status(logger)

        elif choice == "q":
            stop(logger)
            return

        elif choice == "r":
            # Quit, but with non-success code.
            sys.exit(1)

def usage():
    usage = """Usage:  %s <action> [-s]

    Actions:
      status                Show status of all services
      bootstrap             Bootstrap data (re-create data structures)
      deploy                Deploy service(s)
      migrate               Migrate data to latest schema
      provision             Provision/reprovision machine
      undeploy              Undeploy service(s)
      start                 Start service(s)
      stop                  Stop service(s)
      verify                Verify environment
      apply-permissions     Reapply permissions on files and folders
      reinitialize          Reinitialize services

      await-dependencies    Wait for dependencies to become available
      check-ports           Check port availability. Requires -s parameter.
      show-endpoints        Show endpoints
      show-dependencies     Show deployment dependencies

    Parameters:
      -s --service           Service name
      -f --site_options_file Path to site-options.conf (default %s)
         --skip-verification Skip environment verification
      -y --yes               Skip confirmation prompts

    Note: All actions operate on site-options.conf, not runtime-options.conf.

    """ % (os.path.basename(__file__), app.defaults.DEFAULT_SITEOPTIONS_PATH)
    print usage

def main(argv):
    verifyUser()

    #
    # Parse command line
    #

    if len(argv) == 0:
        usage()
        sys.exit(2)

    if argv[0] in ("-h", "--help"):
        usage()
        sys.exit(2)

    action = argv[0]

    try:
        opts, args = getopt.getopt(
            argv[1:],
            "d:s:f:ydVow",
            [
                "service=",
                "site_options_file=",
                "verbose",
                "skip-verification",
                "yes",

                # Below are just kept for compatibility with 16.7 create-runtime.py.
                "force",
                "directory",
                "site_option",
                "noninteractive",
                "await-dependencies"
            ])
    except getopt.GetoptError, err:
        print str(err)
        usage()
        sys.exit(2)

    serviceName = None
    awaitDependencies = False
    siteOptionsPath = app.defaults.DEFAULT_SITEOPTIONS_PATH

    log = app.logutil.getRuntimeLogger()

    for opt, arg in opts:
        if opt in ("-s", "--service"):
            serviceName = arg
        elif opt in ("-f", "--site_options_file"):
            siteOptionsPath = arg
        elif opt in ("--skip-verification"):
            app.servicecontrol.ServiceControlGlobalFlags.verificationEnabled = False
        elif opt in ("-y", "--yes"):
            app.servicecontrol.ServiceControlGlobalFlags.confirmDangerousActions = False
        elif opt in ("-V", "--verbose"):
            app.servicecontrol.ServiceControlGlobalFlags.debugMode = True
            log.setLevel(logging.DEBUG)
        elif opt in ("--force", "-d", "--directory", "-o", "--site_option, --noninteractive"):
            print "%s option is not supported anymore and will be ignored" % opt
        elif opt in ("-w", "--await-dependencies"):
            awaitDependencies = True
        elif opt in ("-h", "--help"):
            usage()
            sys.exit(2)

    # Load landscape information
    app.landscape.Landscape.initializeFromSiteOptionsPath(siteOptionsPath)
    verifyDeploymentLeftovers()

    if app.landscape.Landscape.getSite().getAllClusters():
        print ""
        print app.colors.Color.yellow("WARNING: This site uses clustering. Clustering is an experimental feature that")
        print app.colors.Color.yellow("         is not supposed to be used for production purposes.")
        print ""

    service = None
    if serviceName:
        # Check if service is known
        service = app.landscape.Landscape.getSite().getLocalMachine().getService(serviceName)
        if not service:
            print "Unknown service '%s'" % serviceName
            sys.exit(2)

    # Now, run selected action
    if action == "status":
        status(log, service)
    elif action == "show-endpoints":
        dumpEndpoints(log)
    elif action == "show-dependencies":
        dumpDeploymentDependencies(log)
    elif action == "check-ports":
        checkPorts(log, service)
    elif action == "await-dependencies":
        awaitStartupDependencies(log, service)
    elif action == "verify":
        verifyEnvironment(log, service)
    elif action == "deploy":
        deploy(log, service)
    elif action == "undeploy":
        undeploy(log, service)
    elif action == "bootstrap":
        bootstrap(log, service)
    elif action == "migrate":
        migrate(log, service)
    elif action == "start":
        start(log, service)
    elif action == "stop":
        stop(log, service)
    elif action == "reinitialize":
        reinitialize(log, service)
    elif action == "restart":
        stop(log, service)
        start(log, service)
    elif action == "provision":
        provision(log, awaitDependencies)
    elif action == "apply-permissions":
        applyPermissions(log)
    elif action == "run":
        run(log, awaitDependencies)

    else:
        print "Unknown action specified: %s" % action
        sys.exit(2)


if __name__ == '__main__':
    try:
        main(sys.argv[1:])
    except Exception, err:
        sys.stderr.write(app.colors.Color.red("%s\n" % str(err)))

        if "--verbose" in sys.argv or "-V" in sys.argv:
            sys.stderr.write("\n")
            sys.stderr.write("%s\n" %''.join(app.util.getStackTraceAndException()).strip())

        sys.exit(-1)
