﻿__copyright__ = "Copyright (c) 1999-2013 CollabNet, Inc. All rights reserved."

import sys
import traceback
import app.util

# Initialising the Runtime options configuration
gCfg        = app.util.getRuntimeConfiguration()

# Get the Database Type
gDbType     = gCfg.get('DATABASE_TYPE')

# To enable/disable print error traces on stdout/stderr
gSilentMode = False

def getConnection(dbName=None):
    """
     Return a db connection to the CTF database. If dbName is not specified
     this will connect to the default CTF app database.
     Specifying dbName="" can be used to turn off that behavior.
    """

    global gCfg
    if dbName == None:
       dbName = gCfg.get('DATABASE_NAME')

    return getDBConnection(gCfg.get('DATABASE_TYPE'),
                           gCfg.get('ctfcore-app/dependency/ctfcore-database/jdbc/host'),
                           gCfg.get('ctfcore-app/dependency/ctfcore-database/jdbc/port'),
                           dbName,
                           gCfg.get('DATABASE_USERNAME'),
                           gCfg.get('DATABASE_PASSWORD'))

def getAdminConnection(dbName = None):
    """
     Return a db connection for the postgres user to the CTF database.
     If dbName is not specified this will connect to the default database.
     Specifying dbName="" can be used to turn off that behavior. Note
     the Admin connection is always connect to localhost, with no password.
    """

    global gCfg
    if dbName == None:
       dbName = gCfg.get('DATABASE_NAME')

    if (gCfg.get('DATABASE_TYPE') != 'postgresql'):
        printMessage("ERROR: Admin connection cannot be made for Oracle Database.")
        return None

    return getDBConnection(gCfg.get('DATABASE_TYPE'),
                           None,
                           gCfg.get('ctfcore-app/dependency/ctfcore-datamart/jdbc/port'),
                           None,
                           'postgres',
                           None)

def getReportingConnection(dbName=None):
    """
     Return a db connection to the Reporting database.  If dbName is not
     specified this will connect to the default database.
     Specifying dbName="" can be used to turn off that behavior.
    """

    global gCfg
    if dbName == None:
       dbName = gCfg.get('REPORTS_DATABASE_NAME')

    return getDBConnection(gCfg.get('DATABASE_TYPE'),
                           gCfg.get('ctfcore-app/dependency/ctfcore-datamart/jdbc/host'),
                           gCfg.get('ctfcore-app/dependency/ctfcore-datamart/jdbc/port'),
                           dbName,
                           gCfg.get('REPORTS_DATABASE_USERNAME'),
                           gCfg.get('REPORTS_DATABASE_PASSWORD'))

def getReportingAdminConnection(dbName = None):
    """
     Return a db connection for the admin user to the Reporting database.
     If dbName is not specified this will connect to the default database.
     Specifying dbName="" can be used to turn off that behavior.
    """

    global gCfg
    if dbName == None:
       dbName = gCfg.get('REPORTS_DATABASE_NAME')

    if (gCfg.get('DATABASE_TYPE') != 'postgresql'):
        printMessage("ERROR: Admin connection cannot be made for Oracle Database.")
        return None

    return getDBConnection(gCfg.get('DATABASE_TYPE'),
                           None,
                           gCfg.get('ctfcore-app/dependency/ctfcore-datamart/jdbc/port'),
                           None,
                           'postgres',
                           None)

def getDBConnection(dbType, dbHost, dbPort, dbName, dbUser, dbPassword):
    """
    Returns a database connection using the arguments passed. This is
    required when multiple db server instances are running on the same host.
    """
    con = None
    try:
        if dbType == 'postgresql':
            import psycopg2
            con = psycopg2.connect(user = dbUser,
                               password = dbPassword,
                               database = dbName,
                               host = dbHost,
                               port = dbPort)
            con.autocommit=True
        elif dbType == 'oracle':
            orcPort = gCfg.get('ctfcore-app/dependency/ctfcore-database/jdbc/port')
            import cx_Oracle
            dsn = cx_Oracle.makedsn(dbHost,
                                    orcPort,
                                    dbName)
            con = cx_Oracle.connect(u"%s/%s@%s" \
                        % (dbUser, dbPassword, dsn))
    except StandardError, e:
        printMessage("Reason: %s"%e)
        return None

    return con

def releaseConnection(con):
    """
    Closes or returns to a pool the specified JDBC Connection
    object.
    """
    if con:
        try:
            con.close()
        except:
            pass

def execute(con,sql,bindVals=None):
    """
    Runs a SQL statement which returns no results.
    """
    global gDbType

    try:
        cur = con.cursor()
        runsql = preprareSql(sql, bindVals)
        if (gDbType == 'oracle'):
           cur.execute(u"%s" %runsql)
        else:
           cur.execute(runsql)
        con.commit()
        return cur.rowcount
    except Exception, e:
        obj, inst, trc = sys.exc_info()
        msg = ""
        for i in traceback.format_exception(obj,inst,trc):
            msg = msg + i
        printMessage(msg)
        raise e

def executeUpdate(con,sql,bindVals=None):
    """
    Runs a SQL statement that returns number of rows affected by query
    """
    global gDbType

    try:
        cur = con.cursor()
        runsql = preprareSql(sql, bindVals)
        if (gDbType == 'oracle'):
           cur.execute(u"%s" %runsql)
        else:
           cur.execute(runsql)
        con.commit()
        return cur.rowcount
    except Exception, e:
        obj, inst, trc = sys.exc_info()
        msg = ""
        for i in traceback.format_exception(obj,inst,trc):
            msg = msg + i
        printMessage(msg)
        raise e

def fetchAll(con, sql, bindVals=None):
    """
    Runs a SQL query and returns the results
    """
    global gDbType

    results = []
    try:
        cur = con.cursor()
        runsql = preprareSql(sql, bindVals)
        if (gDbType == 'oracle'):
           cur.execute(u"%s" %runsql)
        else:
           cur.execute(runsql)
        results = cur.fetchall()
        return results
    except Exception, e:
        obj, inst, trc = sys.exc_info()
        msg = ""
        for i in traceback.format_exception(obj,inst,trc):
            msg = msg + i
        printMessage(msg)
        raise e

def preprareSql(sql, bindVals=None):
    """
    Binds values into SQL
    """
    if bindVals != None:
       sql = sql%(bindVals)

    return sql


def run(sql,bindVals=None):
    """
    Runs an SQL query using the database connection against
    the app database.
    """
    execute(getConnection(),sql,bindVals)

def runReporting(sql,bindVals=None):
    """
    Runs an SQL query using the database connection against
    the reporting database.
    """
    execute(getReportingConnection(),sql,bindVals)

def runQuery(sql, bindVals=None):
    """
    Runs an SQL query using the database connection against
    the app database and returns the results.
    """
    return fetchAll(getConnection(),sql, bindVals);

def runReportingQuery(sql, bindVals=None):
    """
    Runs an SQL query using the database connection against
    the reporting database and returns the results.
    """
    return fetchAll(getReportingConnection(),sql, bindVals)

def printMessage(logMsg):
    """
    Print log message to stdout
    """
    global gSilentMode

    if (gSilentMode == False):
       print logMsg
