﻿
__copyright__ = "Copyright (c) 1999-2014 CollabNet, Inc. All rights reserved."
import app.util
import sys
import time
import app.secureconfig
import app.logutil

"""
The module that defines all base classes used for python hook-scripts execution
"""

class BaseScript:

    """The interface defining all CollabNet scripts."""

    def __init__(self, log, cfg, version = None):
        self.cfg = cfg
        self.log = log
        self.version = version

    def logMesg(self, msgType, msgData, lineFeed=True):
        app.logutil.printLog(self.log, msgType, msgData, lineFeed)

    def run(self):
        """
        Main method for invoking the script.
        Should be overridden in the sub classes
        """
        raise AttributeError('Method not implemented')

class BaseOptionScript(BaseScript):
    """
    The base class for all options hook scripts.
    """
    def optionSnippet(self):
        """
        Main method which corresponds to the running the actual python option
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):
        return self.optionSnippet()



class BaseCreateRuntimeScript(BaseScript):
    """
    The base class for all create runtime hook scripts.
    """
    def createRuntime(self):
        """
        Main method which corresponds to the running the actual python option
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):

        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
           sTime = time.time()
           self.createRuntime()
           elapsedTime = time.time() - sTime
           elapsedTime = "{0:.2f}".format(elapsedTime)
           self.logMesg('success', 'Done'.ljust(8), False)
           self.logMesg('info', '%s secs'%elapsedTime.rjust(6))
           return 0
        except Exception, msg:
           self.logMesg('error', 'Failed')
           self.logMesg('error', str(msg))
           self.log.error( str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())
           raise msg

class BasePostInstallScript(BaseScript):
    """
    The base class for all post install hook scripts.
    """
    errorMsg = None
    skipMsg  = None
    confDataList = None
    confDataInfo = None
    reconfigure = False

    def __init__(self, log, version = None):
        return BaseScript.__init__(self, log, app.util.getRuntimeConfiguration(), version)

    def setSkipMsg(self, msg):
        self.skipMsg = msg

    def getSkipMsg(self):
        return self.skipMsg

    def addError(self, msg):
        if self.errorMsg == None:
           self.errorMsg = msg
        else:
           self.errorMsg += msg

    def isError(self):
        if (self.errorMsg == None):
           return False
        return True

    def getError(self):
        return self.errorMsg

    def isApplicable(self):
        """
        Template function to check whether this snippet is applicable to run in
        in the current context.
        """
        return True

    def canRun(self):
        """
        Template function to check whether the environment is conducive to run
        this snippet.
        """
        return True

    def postInstall(self):
        """
        Main method which corresponds to the running the actual python option
        snippet
        """
        raise AttributeError('Method not implemented.')

    def deleteConfigs(self):
        if self.confDataList == None:
           return

        secCfg = app.secureconfig.SecureConfig(self.cfg)

        for sectionConf in self.confDataList:
            for section in sectionConf:
                keys = sectionConf[section]
                for key in keys:
                    secCfg.delConfig(section, key)

    def dataInput(self, force=False):
        if self.confDataList == None:
           return

        import getpass
        secCfg = app.secureconfig.SecureConfig(self.cfg)

        for sectionConf in self.confDataList:
            for section in sectionConf:
                keys = sectionConf[section]
                for key in keys:
                    val = secCfg.getConfig(section, key)
                    if (val == None or force == True):
                       while True:
                          if (key.endswith("secure")):
                             val = getpass.getpass("  %sEnter%s %s : "%(self.INPUT, self.ENDC, self.confDataInfo[key]))
                          else:
                             val = raw_input("  %sEnter%s %s : "%(self.INPUT, self.ENDC, self.confDataInfo[key]))
                          if len(val) == 0:
                             continue
                          else:
                             break
                       secCfg.setConfig(section, key, val)

    def runWithReconfigure(self):

        self.dataInput(False)
        self.reconfigure = True
        return self.run()

    def run(self):
        """
        Exit Codes:
          None: if not applicable
          -1  : if skipped
          0   : if success
          1   : if failed
        """

        if self.isApplicable() != True:
           return None

        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)
        if not self.canRun():
           if self.skipMsg != None:
              self.logMesg('warn', 'Skipped'.ljust(8), False)
              self.logMesg('info', '(%s)'%(self.getSkipMsg()))
           else:
              self.logMesg('warn', 'Skipped')
           return -1

        try:
           sTime = time.time()
           self.postInstall()
           elapsedTime = time.time() - sTime
           elapsedTime = "{0:.2f}".format(elapsedTime)
           self.logMesg('success', 'Done'.ljust(8), False)
           self.logMesg('info', '%s secs'%elapsedTime.rjust(6))
           return 0
        except Exception, msg:
           self.logMesg('error', 'Failed')
           stack_Trace = str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip()
           self.logMesg('debug', stack_Trace)
           if (self.isError()):
              self.logMesg('error', self.getError())
           return 1

class BaseConfigScript(BaseScript):
    """
    The base class for all options hook scripts.
    """
    def configSnippet(self):
        """
        Main method which corresponds to the running the actual python option
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):
        return self.configSnippet()

class BaseBootstrapScript(BaseScript):
    """
    The base class for all bootstrap hook scripts.
    """
    def bootstrap(self):
        """
        Main method which corresponds to the running the actual python bootstrap
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):

        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
           sTime = time.time()
           self.bootstrap()
           elapsedTime = time.time() - sTime
           elapsedTime = "{0:.2f}".format(elapsedTime)
           self.logMesg('success', 'Done'.ljust(8), False)
           self.logMesg('info', '%s secs'%elapsedTime.rjust(6))
           return 0
        except Exception, msg:
           self.logMesg('error', 'Failed')
           self.logMesg('error', str(msg))
           self.log.error( str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())
           raise msg

class BaseMigrationScript(BaseScript):
    """
    The base class for all migration scripts. Each migration script should extend
    this class.
    """
    def migrate(self):
        """
        Main method which corresponds to the running the actual migration activity
        """
        raise AttributeError('Method not implemented. No migration code')

    def preMigrate(self):
        return True

    def postMigrate(self):
        pass

    def run(self):
        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
            sTime = time.time()

            if self.preMigrate():
                self.migrate()
                self.postMigrate()
                self.logMesg('success', 'Done'.ljust(8), False)
            else:
                self.logMesg('success', 'Ignored'.ljust(8), False)

            elapsedTime = time.time() - sTime
            elapsedTime = "{0:.2f}".format(elapsedTime)
            self.logMesg('info', '%s secs'%elapsedTime.rjust(6))

            return 0
        except Exception, msg:
            self.logMesg('error', 'Failed')
            self.logMesg('error', str(msg))
            self.log.error( str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())
            sys.exit(1)

class BaseCronScript(BaseScript):
    """
    The base class for all cron job scripts. Each cron job script should extend
    this class.
    """
    def cronjob(self):
        """
        Main method which corresponds to the running the actual cron job activity
        """
        raise AttributeError('Method not implemented. No cron job code')

    def run(self):
        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
            sTime = time.time()

            self.cronjob()
            self.logMesg('success', 'Done'.ljust(8), False)

            elapsedTime = time.time() - sTime
            elapsedTime = "{0:.2f}".format(elapsedTime)
            self.logMesg('info', '%s secs'%elapsedTime.rjust(6))

            return 0
        except Exception, msg:
            self.logMesg('error', 'Failed')
            self.logMesg('error', str(msg))
            self.log.error( str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())


class BaseVerificationScript(BaseScript):
    """
    The base class for all verification hook scripts.
    """
    def verificationSnippet(self):
        """
        Main method which corresponds to the running the actual python verification
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):

        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
           sTime = time.time()
           self.verificationSnippet()
           elapsedTime = time.time() - sTime
           elapsedTime = "{0:.2f}".format(elapsedTime)
           self.logMesg('success', 'Done'.ljust(8), False)
           self.logMesg('info', '%s secs'%elapsedTime.rjust(6))
           return 0
        except Exception, msg:
           self.logMesg('error', 'Failed')
           self.logMesg('error', str(msg))
           self.log.error( str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())

class BaseInitializationScript(BaseScript):
    """
    The base class for all initialization hook scripts.
    """
    def initialize(self):
        """
        Main method which corresponds to the running the actual python initialization
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):

        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
           sTime = time.time()
           self.initialize()
           elapsedTime = time.time() - sTime
           elapsedTime = "{0:.2f}".format(elapsedTime)
           self.logMesg('success', 'Done'.ljust(8), False)
           self.logMesg('info', '%s secs'%elapsedTime.rjust(6))
           return 0
        except Exception, msg:
           self.logMesg('error', 'Failed: %s' % str(msg) )
           self.log.error(str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())

class BaseFolderScript(BaseScript):
    """
    The base class for all folder hook scripts.
    """
    def deleteFolders(self):
        raise AttributeError('Method not implemented.')

    def createFolders(self):
        raise AttributeError('Method not implemented.')

    def applyFolderPermissions(self):
        raise AttributeError('Method not implemented.')
