#!/usr/bin/python

import getopt, sys, os, re, time, getpass
import urllib
import logging
import app.util
import app.options
import app.logutil
import app.secureconfig
import app.batch
import app.defaults
import app.landscape
from suds.client import Client
from suds.cache import ObjectCache



class PostInstallDataInputBatch(app.batch.Batch):
    def __init__(self, logger, config):
        return super(PostInstallDataInputBatch, self).__init__(
           'PostInstallScript', 'PostInstallScript', 'dataInput', logger, config)

class PostInstallBatch(app.batch.Batch):
    def __init__(self, logger, config):
        return super(PostInstallBatch, self).__init__(
           'PostInstallScript', 'PostInstallScript', 'run', logger, config)

class PostInstallWithReconfigureBatch(app.batch.Batch):
    def __init__(self, logger, config):
        return super(PostInstallWithReconfigureBatch, self).__init__(
           'PostInstallScript', 'PostInstallScript', 'runWithReconfigure', logger, config)

def printLog(log, logType, logMsg, lineFeed=True):
    """ Prints program output messages both to logfile and tty """

    _HEADER = '\033[95m'
    _OKBLUE = '\033[94m'
    _OKGREEN = '\033[92m'
    _WARNING = '\033[93m'
    _FAIL = '\033[91m'
    _ENDC = '\033[0m'

    if logType == "warn":
        display(_WARNING + logMsg + _ENDC, lineFeed)
        log.warn(logMsg)
    elif logType == "error":
        display(_FAIL + logMsg + _ENDC, lineFeed)
        log.error(logMsg)
    elif logType == "success":
        display(_OKGREEN + logMsg + _ENDC, lineFeed)
        log.info(logMsg)
    elif logType == "header":
        display(_HEADER + logMsg + _ENDC, lineFeed)
        log.info(logMsg)
    elif logType == "title":
        display(_OKBLUE + logMsg + _ENDC, lineFeed)
        log.info(logMsg)
    elif logType == "notify" or logType == "info":
        display(logMsg, lineFeed)
        log.info(logMsg)
    elif logType == "debug":
        display(logMsg, lineFeed)
        log.debug(logMsg)
    else:
        display(logMsg)
        log.info(logMsg, lineFeed)

def display(logMsg, lineFeed=True):
    sys.stdout.write(logMsg)
    if (lineFeed):
       sys.stdout.write("\n")
    sys.stdout.flush()


def printUsageAndExit(error=None, exit_code=0):
    """ Prints Help content about this program and exits """

    print
    print "This script does TeamForge post installation tasks."
    print
    print "Valid options: "
    print "  -h | --help         print this usage message and exit"
    print "  -D | --debug        displays all the debug messages"
    print "  -r | --reconfigure  to override the previous post installation parameters"
    print "  -O | --option       pass post-install parameters as arguments to avoid input prompt"
    print "                      valid values: (*) teamforge-tf.siteadmin.username=<USERNAME>"
    print "                                    (*) teamforge-tf.siteadmin.passwd.secure=<PASSWORD>"
    print "                                    (*) gerrit-tf.gerritdb.passwd.secure=<PASSWORD>"
    print
    print
    if error:
        print >> sys.stderr, error
    sys.exit(exit_code)

def runScripts(folder, logger):
    global appCfg, reconfigure, inputCfg

    try:
        for section in inputCfg.iterkeys():
            for key in inputCfg[section].iterkeys():
                value = inputCfg[section][key]
                setConfigParamToFile(section, key, value)
    except:
        #Do nothing. Just proceed with dataInput
        pass

    scriptFolder = os.path.join(app.defaults.DEFAULT_SERVICE_DIR, "legacy-runtime", "hooks", folder)

    # Invoke snippet's dataInput method
    if reconfigure == False:
       PostInstallDataInputBatch(logger, appCfg) \
           .addFolder(scriptFolder) \
           .execute()

       PostInstallBatch(logger, appCfg) \
           .addFolder(scriptFolder) \
           .execute()
    else:
       PostInstallWithReconfigureBatch(logger, appCfg) \
           .addFolder(scriptFolder) \
           .execute()

def getTeamforgeConfig(confKey):
    """ Returns the value of a teamforge configuration token """

    global appCfg

    if not isinstance(appCfg, app.options.RuntimeOptions):
       appCfg = app.util.getRuntimeConfiguration()
    try:
       return appCfg[confKey]
    except:
       return ''

def checkRunProgram():
    """ Check whether this program can run with the current environment. The check that are made are
        Basic/Mandatory Check
            Check if run by root user
            Check if the script is run from /opt/collabnet/teamforge/runtime/scripts
            Check if Teamforge is installed
            Check if Runtime creation is performed
    """
    printLog(logger, "title", "TeamForge Environment Validation");

    printLog(logger, "info", "  Checking user privileges ...".ljust(60), False)
    if (getpass.os.getuid() != 0):
       printLog(logger, "error", "Failed")
       printLog(logger, "error", "Error: This program should run as root user.")
       sys.exit(1)
    printLog(logger, "success", "Passed")

    printLog(logger, "info", "  Verifying Teamforge runtime ...".ljust(60), False)
    startupScript = os.path.normpath("/etc/init.d/collabnet")
    if (not os.path.exists(startupScript)):
       printLog(logger, "error", "Failed")
       printLog(logger, "error", "Error: This program should run only after Teamforge is configured and runtime creation.")
       sys.exit(1)
    printLog(logger, "success", "Passed")

    teamforgeUrl = getTeamforgeConfig("API_SOAP")
    soapUrl = teamforgeUrl + "/ce-soap60/services/CollabNet"

    waitTillJbossIsUp(soapUrl)

    printLog(logger, "info", "  Checking scmadmin login credentials ...".ljust(60), False)
    cnWsdl = "%s?wsdl" %soapUrl
    sudsCacheDir = os.path.join("%s" %getTeamforgeConfig('TEMPORARY_DIR'), "suds")
    sudsCache = ObjectCache(sudsCacheDir)
    sudsLogFile = os.path.join(getTeamforgeConfig('RUNTIME_LOG_DIR'), 'suds.log')
    logging.basicConfig(filename=sudsLogFile)
    logging.getLogger('suds.client').setLevel(logger.getEffectiveLevel())
    collabnetService = Client(cnWsdl, cache=sudsCache, location=cnWsdl)
    try:
        scmadminPswd = getTeamforgeConfig('SCM_ADMIN_PASSWORD')
        sessionId = collabnetService.service.login('scmadmin', scmadminPswd)
        soapLoginStatus = True
    except Exception:
        soapLoginStatus = False

    if soapLoginStatus == True:
       printLog(logger, "success", "Passed")
    else:
       printLog(logger, "warn", "Failed")
       printLog(logger, "warn", "Reason  : Could be scmadmin's password might have been tampered!")
       printLog(logger, "warn", "Solution: Please stop the site and try running Teamforge's install.sh again.")

    printLog(logger, "info", "")

def getResponseObj(soapUrl):
    try:
      urlObj = urllib.urlopen(soapUrl)
    except Exception, err:
       printLog(logger, "error", "Failed")
       printLog(logger, "error", "Error: Unable to connect to TeamForge Application. Could be apache service is down!!")
       printLog(logger, "error", "Reason: [%s]."%str(err))
       sys.exit(1)

    return urlObj

def waitTillJbossIsUp(soapUrl, maxSecondsToWait=5):
    printLog(logger, "info", "  Checking whether TeamForge application is up ...".ljust(60), False)

    number_of_retries = 50
    i = 0
    while i < number_of_retries:
        i += 1
        response_code = getResponseObj(soapUrl).getcode()
        if response_code == 500:
            printLog(logger, "error", "500 Internal Server Error")
            printLog(logger, "error", "Server encountered an unexpected condition which prevented it from fulfilling the request")
            sys.exit(1)
        elif response_code == 200:
            break
        elif response_code != 200:
            time.sleep(maxSecondsToWait)

    if i == number_of_retries:
        printLog(logger, "error", "Failed")
        printLog(logger, "error", "Error: TeamForge Application is not up and running. Start the TeamForge application.")
        printLog(logger, "error", "Reason: Not received successful http response while accessing [%s]."%soapUrl)
        sys.exit(1)

    printLog(logger, "success", "Passed")


def main():
    statuappList = {}
    printLog(logger, "title", "TeamForge Post Installation");
    statuappList = runScripts("post-install", logger)
    printLog(logger, "info", "")

    elapsed = time.time() - startTimeStamp
    elapsed = "{0:.2f}".format(elapsed)
    printLog(logger, "title", "Post Installation Summary");

    printLog(logger, "info", "  Total time taken".ljust(50) + " : ", False)
    printLog(logger, "info", "%s seconds" % elapsed)

def setConfigParamToFile(section, param, value):
    try:
      secCfg  = app.secureconfig.SecureConfig(appCfg)
      secCfg.setConfig(section, param, value)
    except Exception, e:
      return False

    return True

def preparePostInstallConfiguration():
    setConfigParamToFile('teamforge', 'tf.siteadmin.username', 'scmadmin')

    scmadminPswd = getTeamforgeConfig('SCM_ADMIN_PASSWORD')
    if len(scmadminPswd):
      setConfigParamToFile('teamforge', 'tf.siteadmin.passwd.secure', scmadminPswd)

    gerritDBPswd = getTeamforgeConfig('GERRIT_DATABASE_PASSWORD')
    if len(gerritDBPswd):
      setConfigParamToFile('gerrit', 'tf.gerritdb.passwd.secure', gerritDBPswd)

def setConfigParam(inStr):
    global inputCfg

    try:
      mObj = re.match('^(\w+)-([\w\.]+)=(.*)$', inStr)
      if not mObj:
         return False
      section =  mObj.groups()[0]
      param   =  mObj.groups()[1]
      value   =  mObj.groups()[2]
      if not inputCfg.has_key(section):
         inputCfg[section] = {}
      inputCfg[section][param] = value
    except Exception, e:
      return False

    return True

def configureSsl():
    if os.environ.get("POSTINSTALL_DISABLE_SSL_VALIDATION", "false") == "true":
        import ssl
        try:
            _create_unverified_https_context = ssl._create_unverified_context
        except AttributeError:
            # Legacy Python that doesn't verify HTTPS certificates by default
            pass
        else:
            # Handle target environment that doesn't support HTTPS verification
            ssl._create_default_https_context = _create_unverified_https_context

## Script's Main enterance Block
if __name__ == "__main__":
    app.landscape.Landscape.initializeFromSiteOptionsPath(app.defaults.RUNTIME_OPTIONS_CONF)
    appCfg     = app.util.getRuntimeConfiguration()

    siteDir = appCfg['SITE_DIR']
    runtimeDir      = appCfg['RUNTIME_DIR']
    distributionDir = appCfg['DISTRIBUTION_DIR']
    startTimeStamp  = time.time()

    debug       = False
    reconfigure = False
    logger      = None
    inputCfg = {}
    optionList = []

    configureSsl()

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hDrO:",
                                   ["help", "debug", "reconfigure", "option"])
    except getopt.GetoptError:
        printUsageAndExit("Invalid options specified!!\n", 1)
    for option, value in opts:
        if option in ("-h", "--help"):
            printUsageAndExit()
        elif option in ('-D', "--debug"):
            debug = True
        elif option in ('-r', "--reconfigure"):
            reconfigure = True
        elif option in ('-O', "--option"):
            setConfigParam(value)

    runtime_logDir = appCfg['RUNTIME_LOG_DIR']
    logger = app.logutil.getPostInstallLogger(runtime_logDir, False)
    if debug:
       logger.setLevel(10)
    else:
       logger.setLevel(20)

    postInstallLog = os.path.normpath(os.path.join(runtime_logDir, 'postinstall.log'))

    printLog(logger, "header", "TeamForge Post Installation".center(78))
    checkRunProgram()
    preparePostInstallConfiguration()

    try:
        main()
    except Exception, err:
        logger.error(''.join(app.util.getStackTraceAndException()).strip())

        printLog(logger, "error", "\n\n")
        printLog(logger, "error", "Post Installation failed: %s" % str(err))
        printLog(logger, "error", "\n\n")
        printLog(logger, "error", "See %s for details" % postInstallLog)
        sys.exit(-1)

